/*
	STM32F4
	WS2812B control with SPI & DMA, PLL in use, SysTick-driven animation
	gbm, 11'2015
*/
#include "board.h"
#include "WS2812.h"
//========================================================================
#define SYSCLK_FREQ	80000000u

extern void app_init(void);
void WEAK app_init(void) {}

int main(void)
{
	// PLL - 80 MHz
	RCC->PLLCFGR = (RCC->PLLCFGR & RCC_PLLCFGR_RSVD)
		| RCC_PLLCFGR_PLLQV(8) | RCC_PLLCFGR_PLLPV(4)
		| RCC_PLLCFGR_PLLNV(160) | RCC_PLLCFGR_PLLMV(8);
	RCC->CR |= RCC_CR_PLLON;	//
	RCC->CFGR = RCC_CFGR_PPRE2_DIV2 | RCC_CFGR_PPRE1_DIV2;	// APB2, APB1 prescaler = 2
	// set Flash speed
	FLASH->ACR = FLASH_ACR_PRFTEN | FLASH_ACR_LATENCY_2WS;	// 2ws 64..90
	while (!(RCC->CR & RCC_CR_PLLRDY));
	RCC->CFGR |= RCC_CFGR_SW_PLL;
	while ((RCC->CFGR & RCC_CFGR_SWS) != RCC_CFGR_SWS_PLL);

	// enable peripherals
	RCC->AHB1ENR = RCC_AHB1ENR_GPIOAEN | RCC_AHB1ENR_GPIOBEN
		| RCC_AHB1ENR_GPIOCEN | RCC_AHB1ENR_DMA2EN;
	RCC->APB1ENR = RCC_APB1ENR_PWREN;
	RCC->APB2ENR = RCC_APB2ENR_SPI1EN;
	// port setup - SPI1 MOSI at PA7 and PB5
	GPIOA->AFR[0] = BF4(7, 5);	// SPI1: AF5
	GPIOB->AFR[0] = BF4(5, 5);	// SPI1: AF5
	GPIOA->OSPEEDR = GPIOA_OSPEEDR_SWD | BF2(7, GPIO_OSPEEDR_MED);	// MOSI - medium speed
	GPIOA->MODER = GPIOA_MODER_SWD | BF2(7, GPIO_MODER_AF);	// MOSI as AF
	GPIOB->OSPEEDR = BF2(5, GPIO_OSPEEDR_MED);	// MOSI - medium speed
	GPIOB->MODER = BF2(5, GPIO_MODER_AF);	// MOSI as AF
	// SPI setup
	SPI1->CR2 = SPI_CR2_TXDMAEN;
	SPI1->CR1 = SPI_CR1_DFF | SPI_CR1_SSM | SPI_CR1_SSI | SPI_CR1_SPE | SPI_CR1_BRDIV16 | SPI_CR1_MSTR;
	// Flash unlock
	FLASH->KEYR = FLASH_FKEY1;
	FLASH->KEYR = FLASH_FKEY2;
	
	app_init();
	PWR->CR = PWR_CR_PLSV(6) | PWR_CR_PVDE;	// PVD: 2.8 V, enable
	while (PWR->CSR & PWR_CSR_PVDO);
	// interrupts and sleep
	EXTI->RTSR = EXTI_MR_PVD;	// PVD
	EXTI->IMR = EXTI_MR_PVD;	// PVD
	NVIC_EnableIRQ(PVD_IRQn);
	SysTick_Config(SYSCLK_FREQ/100);
	SCB->SCR = SCB_SCR_SLEEPONEXIT_Msk;
	__WFI();	// sleep
}
//========================================================================
// WS2812 reset pulse width
// (50 us -> 120 bits @2.4 MHz, 125 bits @2.5 MHz)
#define WS_RST_FRM	8 // no. of 16-bit frames for RESET pulse
//========================================================================
#define NBYTES	(NPIXELS * 3)
// no. of 16-bit frames for LED string control (9 frames per pixel pair)
#define NFRAMES	(WS_RST_FRM + (NPIXELS + 1) / 2 * 9)

void WS2812_start(struct wspix_ *ptr)
{
	static uint16_t enc_wsdata[NFRAMES];

	static const uint16_t encode[] = {
		04444, 04446, 04464, 04466, 04644, 04646, 04664, 04666,
		06444, 06446, 06464, 06466, 06644, 06646, 06664, 06666
	};	// bit-to triple encoding table - 4->12 bits
	
	if (DMA2->LISR & DMA_LISR_TCIF3 || DMA2_Stream3->CR == 0)
	{
		uint8_t *wsptr = (uint8_t *)ptr;
		uint16_t *ep = enc_wsdata + WS_RST_FRM;	// skip reset frames
		uint32_t ev;
		
		for (uint32_t i = 0; i < NBYTES; i ++)
		{
			// encode each 2 bytes as 3 16-bit words
			ev = encode[*wsptr >> 4] << 12;
			ev |= encode[*wsptr ++ & 0xf];	// 24 bits in ev
			*ep ++ = ev >> 8;
			ev <<= 8;	// 8 bits left, left-align at 16-bit boundary
			if (++ i < NBYTES)	// not the last byte
			{
				ev <<= 4;	// make place for 12 bits
				ev |= encode[*wsptr >> 4];	// 20 b total
				*ep ++ = ev >> 4;
				ev <<= 12;	// 4 bits left
				ev |= encode[*wsptr ++ & 0xf];	// 16 bits
			}
			*ep ++ = ev;
		}
		// setup WS2812 data transfer - DMA2 Stream3 ch3
		DMA2_Stream3->CR = 0;	// disable
		DMA2->LIFCR = DMA_LIFCR_CALLF3 | DMA_LIFCR_CALLF2;	// clear stream4 flags
		DMA2_Stream3->PAR = (uint32_t)&SPI1->DR;
		DMA2_Stream3->M0AR = (uint32_t)enc_wsdata;
		DMA2_Stream3->NDTR = NFRAMES;
		// medium_priority, increment memory adress, mem->periph, enable
		DMA2_Stream3->CR = DMA_SxCR_CHSELV(3) | DMA_SxCR_PL_0 | DMA_SxCR_MSIZE16 | DMA_SxCR_PSIZE16
			| DMA_SxCR_MINC | DMA_SxCR_DIR_M2P | DMA_SxCR_EN;
		// clear buffer using Stream2
		DMA2_Stream2->CR = 0;	// disable
		DMA2_Stream2->PAR = (uint32_t)enc_wsdata;	// constant zero
		DMA2_Stream2->M0AR = (uint32_t)ptr;	// buf to clear
		DMA2_Stream2->NDTR = (NBYTES + 3) >> 2;
		// increment memory adress, mem->mem, enable
		DMA2_Stream2->CR = DMA_SxCR_MSIZE32 | DMA_SxCR_PSIZE32
			| DMA_SxCR_MINC | DMA_SxCR_DIR_M2M | DMA_SxCR_EN;
	}
}
