/*
 * WS2812B strip animations
 * gbm, 11'2015
*/
#include "board.h"
#include "WS2812.h"

// functions defined in device-specific Flash module
uint16_t flash_getmode(void);
void flash_storemode(uint16_t m);

// pixel data
static struct wspix_ wsdata[NPIXELS];
//========================================================================
// animation data
#define NDOTS (NPIXELS / 10)
#define DOT_GEN_DELAY		50	// min. dot generation spacing
#define DOT_TWUP_SPEED	4	//5
#define DOT_TWDN_SPEED	2	//15
#define DOT_STEADY_TIME	150

#define NWORMS (NPIXELS / 12)
#define	CLIMB_SPEED	4	// microunits (1/256 pixel) per time unit
#define WORM_GEN_MIN	300	// min. worm generation delay
#define	WORM_GEN_SPAN	((NPIXELS * 256 / CLIMB_SPEED / NWORMS - WORM_GEN_MIN) * 2)	// max. random gen. delay
#define FSCALE	2048
// fireworks flight time - 0 to top
// MAXFTIME^2 / FSCALE = (NPIXELS - 1)
#if		NPIXELS == 60
#define MAXFTIME 348
#elif	NPIXELS == 144
#define MAXFTIME 543
#elif	NPIXELS == 288
#define MAXFTIME 767
#elif	NPIXELS == 432
#define MAXFTIME 940
#elif	NPIXELS == 576
#define MAXFTIME 1087
#endif
#define MINFTIME (MAXFTIME - 131)
#define FTDELTA	33	// diff 131 = 4 * 33 - 1
//========================================================================
static enum mode_ {M_MIXED, M_BLINK, M_TWINKLE, M_FIREWORKS, M_WORMS, M_NMODES} mode = M_BLINK, prev_mode = M_BLINK;
//========================================================================
void app_init(void)
{
	if ((prev_mode = (enum mode_)flash_getmode()) < M_NMODES)
		mode = prev_mode;
}
//========================================================================
// save mode on power down
void PVD_IRQHandler(void)
{
	EXTI->PR = EXTI_MR_PVD;	// clear interrupt rq
	if (mode != prev_mode)		// mode changed - write to Flash
		flash_storemode(prev_mode = mode);
}
//========================================================================
// 31-bit lfsr pseudo-random generator, gbm 01'15
static uint32_t lfsr = 1;
static uint32_t lfsr_rand(void)
{
	if (lfsr & 1)
		lfsr ^= 0x40000004u << 1;
	return lfsr >>= 1;
}
#define rand lfsr_rand
//========================================================================
// occupancy handling for dot objects
static uint32_t pixactive[(NPIXELS + 31) >> 5];

static inline _Bool px_isactive(uint32_t px)
{
	return px < NPIXELS && (pixactive[px >> 5] >> (px & 0x1f) & 1);
}

static inline void px_setactive(uint32_t px)
{
	if (px < NPIXELS)
		pixactive[px >> 5] |= 1 << (px & 0x1f);
}

static inline void px_setinactive(uint32_t px)
{
	if (px < NPIXELS)
		pixactive[px >> 5] &= ~(1 << (px & 0x1f));
}
// get random dot position with no overlap and no neighbour
static uint32_t rand_position(void)
{
	uint32_t pos = rand() % NPIXELS;
	while (px_isactive(pos - 1) || px_isactive(pos) || px_isactive(pos + 1))
		pos = (pos + 3) % NPIXELS;
	px_setactive(pos);
	return pos;
}
//========================================================================
// moving object
enum wmode_ {WM_FREE, WM_CLIMB, WM_THROW, WM_FALL};
static struct worm_ {
	int32_t bias;	// position bias for non-constant speed worms
	int32_t head;	// integer position
	int32_t time;	// object local time
	uint8_t frac;	// fractional part of position
	enum wmode_ mode;
	struct wspix_ pixel[LED_MAX];
} worm[NWORMS];

// fixed dot
enum dotmode_ {DM_FREE, DM_BLINKUP, DM_STEADY, DM_BLINKDOWN, DM_TWINKLE, DM_TWINKLEOFF};
static struct dot_ {
	uint32_t time;
	uint16_t pos;
	uint8_t	val;
	uint8_t	target;
	enum dotmode_ mode;
	struct wspix_ color;
} dot[NDOTS];
//========================================================================
static uint8_t dither = 0;
// accumulate fractional pixel color	
static void acc_pixel(int32_t sp, struct wspix_ *cp, uint32_t fr)
{
	if (sp >= 0 && sp < NPIXELS && fr)
	{
		wsdata[sp].green += (cp->green * fr + dither) / 256;
		wsdata[sp].red += (cp->red * fr + dither) / 256;
		wsdata[sp].blue += (cp->blue * fr + dither) / 256;
	}
}
// accumulate pixel color	
static void acc_full_pixel(int32_t sp, struct wspix_ *cp)
{
	if (sp >= 0 && sp < NPIXELS)
	{
		wsdata[sp].green += cp->green;
		wsdata[sp].red += cp->red;
		wsdata[sp].blue += cp->blue;
	}
}
//========================================================================
void SysTick_Handler(void)
{
	static uint32_t time = 0;
	static uint16_t newworm_timer = 0;
	static uint8_t nxtcol = 0, nxtworm = 0;
	static uint8_t active_worms = 0;
	static uint8_t active_dots = 0;
	int32_t i, j;
	static uint8_t key_state;
	//static const struct wspix_ cc = {.green = 14, .red = 16, .blue = 2};
	
	if ((key_state = key_state << 1 | BTN_PRESSED) == 1
		&& ++ mode == M_NMODES)
			mode = M_MIXED;

	time ++;
	dither ^= 1;

	if (newworm_timer)
		newworm_timer --;
	if ((mode <= M_TWINKLE && active_dots < NDOTS)
		|| ((mode == M_MIXED || mode >= M_FIREWORKS) && active_worms < NWORMS && newworm_timer == 0))
	{
		// Create new object
		// make new color
		struct wspix_ newc;
		switch (nxtcol ++)
		{
		case 1:	// R-G
			newc.red = rand() % LED_MAX;
			newc.green = LED_MAX - newc.red;
			newc.blue = 0;
			break;
		case 3:	// G-B
			newc.red = 0;
			newc.green = rand() % LED_MAX;
			newc.blue = LED_MAX - newc.green;
			break;
		case 5:	// B-R
			newc.green = 0;
			newc.blue = rand() % LED_MAX;
			newc.red = LED_MAX - newc.blue;
			nxtcol = 0;
			break;
		default:	// any color
			newc.red = rand() % LED_MAX;
			newc.green = rand() % (LED_MAX - newc.red);
			newc.blue = LED_MAX - newc.red - newc.green;
		}
		
		if (mode <= M_TWINKLE && active_dots < NDOTS)
		{
			// create dot
			static uint8_t ndot = 0;
			
			active_dots ++;
			while (dot[ndot].mode != WM_FREE)
				if (++ ndot == NDOTS)
					ndot = 0;
			dot[ndot].pos = rand_position();
			dot[ndot].time = 0;
			dot[ndot].val = 0;
			dot[ndot].target = 0;
			dot[ndot].color = newc;
				
			enum mode_ dot_mode = mode;
			if (mode == M_MIXED)
				dot_mode = ndot & 1 ? M_BLINK : M_TWINKLE;

			switch (dot_mode)
			{
				case M_BLINK:
					dot[ndot].mode = DM_BLINKUP;
					break;
				case M_TWINKLE:
//					dot[ndot].color.green = 14;
//					dot[ndot].color.red = 16;
//					dot[ndot].color.blue = 2;
					dot[ndot].mode = DM_TWINKLE;
					break;
			}
		}
		else if (active_worms < NWORMS && (mode == M_MIXED || mode >= M_FIREWORKS))
		{
			// create worm
			static uint16_t newftime = MAXFTIME;
			active_worms ++;
			while (worm[nxtworm].mode != WM_FREE)
				if (++ nxtworm == NWORMS)
					nxtworm = 0;
			j = nxtworm;
			worm[j].frac = 0;
			worm[j].time = 0;
			worm[j].head = 0;
			worm[j].pixel[0] = newc;
				
			enum mode_ worm_mode = mode;
			if (mode == M_MIXED)
				worm_mode = rand() & 1 ? M_WORMS : M_FIREWORKS;

			switch (worm_mode)
			{
			case M_FIREWORKS:
				newworm_timer = WORM_GEN_MIN + rand() % WORM_GEN_SPAN;
				worm[j].mode = WM_THROW;
				if ((newftime += FTDELTA) > MAXFTIME)
				{
					worm[j].time = - MAXFTIME;
					newftime -= MAXFTIME - MINFTIME;
				}
				else
					worm[j].time = - newftime;
				worm[j].bias = worm[j].time * worm[j].time;
				break;
			case M_WORMS:
				newworm_timer = WORM_GEN_MIN + rand() % WORM_GEN_SPAN;
				worm[j].mode = WM_CLIMB;
				break;
			}
		}
	}
	// Animate dots
	for (j = 0; j < NDOTS; j ++)
	{
		if (dot[j].mode != DM_FREE)
		{
			int32_t fr;
			dot[j].time ++;
			switch (dot[j].mode)
			{
			case DM_BLINKUP:
				fr = dot[j].time * DOT_TWUP_SPEED;
				if (fr > 255)
				{
					fr = 255;
					dot[j].time = rand() % DOT_STEADY_TIME;
					dot[j].mode = DM_STEADY;
				}
				dot[j].val = fr;
				break;
			case DM_STEADY:
				if (dot[j].time == DOT_STEADY_TIME)
				{
					dot[j].time = 0;
					dot[j].mode = DM_BLINKDOWN;
				}
				break;
			case DM_BLINKDOWN:
				fr = 255 - (int32_t)dot[j].time * DOT_TWDN_SPEED;
				if (fr <= 0)
				{
					px_setinactive(dot[j].pos);
					dot[j].mode = DM_FREE;
					active_dots --;
				}
				else
					dot[j].val = fr;
				break;
			case DM_TWINKLE:
				if ((dot[j].time & 7) == 0)
				{
					// 12.5 Hz update
					fr = rand() & 0xff;
					if (fr == 0)
					{
						dot[j].mode = DM_TWINKLEOFF;
						dot[j].target = 0;
					}
					else if (((fr &= 0x1f) & 0xc) != 0)
					{
						if (fr > 15)
							fr = 15;
						dot[j].target = fr << 4;	// target intensity - 16 levels
					}
				}
			case DM_TWINKLEOFF:
				if (dot[j].val != dot[j].target)
					dot[j].val = dot[j].val < dot[j].target
						? dot[j].val + 8 : dot[j].val - 8;
				if (dot[j].mode == DM_TWINKLEOFF && dot[j].val == 0)
				{
					px_setinactive(dot[j].pos);
					dot[j].mode = DM_FREE;
					active_dots --;
				}
				break;
			default:
				;
			}
		}
	}
	// Animate existing worms
	for (j = 0; j < NWORMS; j ++)
	{
		if (worm[j].mode != WM_FREE)
		{
			int oldhead, head;

			worm[j].time ++;
			switch (worm[j].mode)
			{
			case WM_THROW:
				oldhead = worm[j].head;
				head = worm[j].bias - worm[j].time * worm[j].time;
				worm[j].head = head / FSCALE;
				worm[j].frac = head % FSCALE / 8;

				if (worm[j].head != oldhead)
				{
					if (worm[j].head < oldhead)
					{
						worm[j].mode = WM_FALL;
						break;
					}
					for (i = LED_MAX - 2; i >= 0; i --)
						worm[j].pixel[i + 1] = worm[j].pixel[i];
				}
				for (i = 1; i < LED_MAX; i ++)
				{
					if (worm[j].pixel[i].green)
						worm[j].pixel[i].green --;
					if (worm[j].pixel[i].red)
						worm[j].pixel[i].red --;
					if (worm[j].pixel[i].blue)
						worm[j].pixel[i].blue --;
				}
				break;
			case WM_FALL:
				oldhead = worm[j].head;
				head = worm[j].bias - worm[j].time * worm[j].time;
				worm[j].head = head / FSCALE;
				worm[j].frac = head % FSCALE / 8;
				if (worm[j].head != oldhead)
				{
					if (worm[j].head == - LED_MAX)
					{
						worm[j].mode = WM_FREE;
						active_worms --;
						break;
					}
					for (i = LED_MAX - 2; i >= 0; i --)
						worm[j].pixel[i + 1] = worm[j].pixel[i];

				}
				for (i = 1; i < LED_MAX; i ++)
				{
					if (worm[j].pixel[i].green)
						worm[j].pixel[i].green --;
					if (worm[j].pixel[i].red)
						worm[j].pixel[i].red --;
					if (worm[j].pixel[i].blue)
						worm[j].pixel[i].blue --;
				}
				break;
			case WM_CLIMB:
				head = worm[j].time * CLIMB_SPEED;
				if (head >= NPIXELS * 256)
				{
					worm[j].mode = WM_FALL;
					worm[j].time = 0;
					worm[j].bias = (NPIXELS - 1) * FSCALE;
					worm[j].head = NPIXELS - 1;
					worm[j].frac = 255;
					for (i = 1; i < LED_MAX; i ++)
					{
						worm[j].pixel[i].green = 0;
						worm[j].pixel[i].red = 0;
						worm[j].pixel[i].blue = 0;
					}
				}
				else
				{
					worm[j].head = head / 256;
					worm[j].frac = head % 256;
				}
				break;
			default:
				;
			}
		}
	}
	// Rasterize dots
	for (i = 0; i < NDOTS; i ++)
		if (dot[i].mode != DM_FREE)
			acc_pixel(dot[i].pos, &dot[i].color, dot[i].val);
	// Rasterize worms
	for (i = 0; i < NWORMS; i ++)
	{
		switch (worm[i].mode)
		{
		case WM_THROW:
			// antialiasing while moving up
			acc_pixel(worm[i].head, &worm[i].pixel[0], worm[i].frac);
			for (j = 1; j < LED_MAX; j ++)
				acc_full_pixel(worm[i].head - j, &worm[i].pixel[j]);
			break;
		case WM_FALL:
			acc_pixel(worm[i].head, &worm[i].pixel[0], 256 - worm[i].frac);
			for (j = 1; j < LED_MAX; j ++)
				acc_full_pixel(worm[i].head + j, &worm[i].pixel[j]);
			break;
		case WM_CLIMB:
			acc_pixel(worm[i].head, &worm[i].pixel[0], worm[i].frac);
			acc_pixel(worm[i].head - 1, &worm[i].pixel[0], 256 - worm[i].frac);
			break;
		default:
			;
		}
	}

	WS2812_start(wsdata);
}
